from henson import Abort, Application

class FileConsumer:
    """Read lines from a file."""

    def __init__(self, filename):
        self.filename = filename
        self._file = None

    def __iter__(self):
        """FileConsumer objects are iterators."""
        return self

    def __next__(self):
        """Return the next line of the file, if available."""
        if not self._file:
            self._file = open(self.filename)
        try:
            return next(self._file)
        except StopIteration:
            self._file.close()
            raise Abort('Reached end of file', None)

    async def read(self):
        """Return the next line in the file."""
        return next(self)

async def callback(app, message):
    """Print the message retrieved from the file consumer."""
    print(app.name, 'received:', message)
    return message

app = Application(
    __name__,
    callback=callback,
    consumer=FileConsumer(__file__),
)

@app.startup
async def print_header(app):
    """Print a header for the file being processed."""
    print('# Begin processing', app.consumer.filename)

@app.teardown
async def print_footer(app):
    """Print a footer for the file being processed."""
    print('# Done processing', app.consumer.filename)

@app.message_preprocessor
async def remove_comments(app, line):
    """Abort processing of comments (lines that start with #)."""
    if line.strip().startswith('#'):
        raise Abort('Line is a comment', line)
    return line



"""Collection of Henson CLI tasks."""

import asyncio
from importlib import find_loader, import_module
import os
import sys
from threading import Thread

from argh import ArghParser, CommandError
from watchdog.events import PatternMatchingEventHandler
from watchdog.observers import Observer

from . import __version__
from .base import Application


def run(application_path, reloader=False, workers=1, debug=False):
    """Import and run an application.

    Args:
        application_path (str): The path to the application that should
            be imported and run.
        reloader (Optional[bool]): Whether or not to watch for changes
            within an application and reload the running instance when
            they are detected. Not recommended for production use.
            Defaults to False.
        workers (Optional[int]): How many async workers the application
            should use to process messages. Defaults to 1.
        debug (Optional[bool]): Whether or not to run the application
            with debug mode enabled. Defaults to False.
    """
    # Add the present working directory to the import path so that
    # services can be found without installing them to site-packages
    # or modifying PYTHONPATH
    sys.path.insert(0, '.')

    # First, find the module that should be imported
    application_path_parts = application_path.split(':', 1)
    import_path = application_path_parts.pop(0)

    # Then, try to find an import loader for the import_path
    # NOTE: this is to handle the case where a module is found but not
    # importable because of dependency import errors (Python 3 only)
    if not find_loader(import_path):
        raise CommandError(
            'Unable to find an import loader for {}.'.format(import_path),
        )

    # Once found, import the module and handle any dependency errors
    # TODO: Wrap the ImportError raised here to provide more meaningful
    # error messages to the end user
    module = import_module(import_path)

    # If an application name is specified, use that to select the
    # application instance
    try:
        app_name = application_path_parts.pop()
        # TODO: Wrap the AttributeError raised here to provide more
        # meaningful error messages to the end user
        app = getattr(module, app_name)
        # If the attribute specified by app_name is a callable, assume
        # it is an application factory and call it to get an instance of
        # a Henson application.
        if callable(app):
            app = app()
        # Fail if the attribute specified is not a Henson application
        if not isinstance(app, Application):
            raise CommandError(
                'app must be an instance of a Henson application. '
                'Got {}'.format(type(app)),
            )

    # If no application name is specified, try to automatically select
    # the correct module attribute based on type
    except IndexError:
        app_candidates = []
        for name in dir(module):
            attr = getattr(module, name)
            if isinstance(attr, Application):
                app_candidates.append((name, attr))

        # If there are zero app_candidates, there's nothing to run.
        if not app_candidates:
            raise CommandError(
                'No Henson application found. Please specify the '
                'application by name or run a different module.',
            )

        # If there are more than one, the choice of which app to run is
        # ambiguous.
        if len(app_candidates) > 1:
            raise CommandError(
                'More than one Henson application found in {}. Please '
                'specify a application by name (probably one of [{}]).'.format(
                    import_path, ', '.join(ac[0] for ac in app_candidates)),
            )

        app_name, app = app_candidates[0]

    if reloader:
        # If the reloader is requested, create threads for running the
        # application and watching the file system for changes
        print('Running {}.{} with reloader...'.format(
            import_path,
            app_name,
        ))

        # Find the root of the application and watch for changes
        watchdir = os.path.abspath(module.__file__)
        for _ in import_path.split('.'):
            watchdir = os.path.dirname(watchdir)

        # Create observer and runner threads
        observer = Observer()
        loop = asyncio.new_event_loop()
        runner = Thread(
            target=app.run_forever,
            kwargs={'num_workers': workers, 'loop': loop},
        )

        # This function is called by watchdog event handler when changes
        # are detected by the observers
        def restart_process(event):
            """Restart the process in-place."""
            os.execv(sys.executable, [sys.executable] + sys.argv[:])

        # Create the handler and watch the files
        handler = PatternMatchingEventHandler(
            patterns=['*.py', '*.ini'],
            ignore_directories=True,
        )
        handler.on_any_event = restart_process
        observer.schedule(handler, watchdir, recursive=True)

        # Start running everything
        runner.start()
        observer.start()

    else:
        # If the reloader is not needed, avoid the overhead
        print('Running {}.{} forever ...'.format(import_path, app_name))
        app.run_forever(num_workers=workers)


def main():
    """Dispatch the CLI command to the target function."""
    return parser.dispatch()


# Define a parser and add commands to it.
parser = ArghParser()
parser.add_argument('--version', action='version', version=__version__)
parser.add_commands([run])



"""Henson."""

import os as _os
import pkg_resources as _pkg_resources

from .base import Application  # NOQA
from .exceptions import Abort  # NOQA
from .extensions import Extension  # NOQA

try:
    _dist = _pkg_resources.get_distribution(__package__)
    if not __file__.startswith(_os.path.join(_dist.location, __package__)):
        # Manually raise the exception if there is a distribution but
        # it's installed from elsewhere.
        raise _pkg_resources.DistributionNotFound
except _pkg_resources.DistributionNotFound:
    __version__ = 'development'
else:
    __version__ = _dist.version



"""Custom exceptions used by Henson."""

__all__ = ('Abort',)


class Abort(Exception):
    """An exception that signals to Henson to stop processing a message.

    When this exception is caught by Henson it will immediately stop
    processing the message. None of the remaining callbacks will be
    called.

    If the exception is caught while processing a result, that result
    will no longer be processed. Any other results generated by the same
    message will still be processed.

    Args:
        reason (str): The reason the message is being aborted. It should
            be in the form of "noun.verb" (e.g., "provider.ignored").
        message: The message that is being aborted. Usually this will be
            the incoming message, but it can also be the result.
    """

    def __init__(self, reason, message):
        """Initialize the class."""
        super().__init__(reason)
        self.message = message



"""Implementation of the service."""

import asyncio
from copy import deepcopy
import logging
import sys
import traceback

from .config import Config
from .exceptions import Abort

__all__ = ('Application',)


class Application:
    """A service application.

    Each message received from the consumer will be passed to the
    callback.

    Args:
        name (str): The name of the application.
        settings (Optional[object]): An object with attributed-based
            settings.
        consumer (optional): Any object that is an iterator or an
            iterable and yields instances of any type that is supported
            by ``callback``. While this isn't required, it must be
            provided before the application can be run.
        callback (Optional[asyncio.coroutine]): A callable object that
            takes two arguments, an instance of
            :class:`henson.base.Application` and the (possibly)
            preprocessed incoming message.  While this isn't required,
            it must be provided before the application can be run.
    """

    def __init__(self, name, settings=None, *, consumer=None, callback=None):
        """Initialize the class."""
        self.name = name

        # Configuration
        self.settings = Config()
        self.settings.from_object(settings or {})
        self.settings.setdefault('DEBUG', False)
        self.settings.setdefault('SLEEP_TIME', 0.1)

        # Callbacks
        self.callback = callback
        self._callbacks = {
            'error': [],
            'message_acknowledgement': [],
            'message_preprocessor': [],
            'result_postprocessor': [],
            'startup': [],
            'teardown': [],
        }

        self.consumer = consumer

        self.logger = logging.getLogger(self.name)

    def error(self, callback):
        """Register an error callback.

        Args:
            callback (asyncio.coroutine): A callable object that takes
                three arguments: an instance of
                :class:`henson.base.Application`, the incoming message,
                and the exception that was raised. It will be called any
                time there is an exception while reading a message from
                the queue.

        Returns:
            asyncio.coroutine: The callback.

        Raises:
            TypeError: If the callback isn't a coroutine.
        """
        self._register_callback(callback, 'error')
        return callback

    def message_acknowledgement(self, callback):
        """Register a message acknowledgement callback.

        Args:
            callback (asyncio.coroutine): A callable object that takes
                two arguments: an instance of
                :class:`henson.base.Application` and the original
                incoming message as its only argument. It will be called
                once a message has been fully processed.

        Returns:
            asyncio.coroutine: The callback.

        Raises:
            TypeError: If the callback isn't a coroutine.
        """
        self._register_callback(callback, 'message_acknowledgement')
        return callback

    def message_preprocessor(self, callback):
        """Register a message preprocessing callback.

        Args:
            callback (asyncio.coroutine): A callable object that takes
                two arguments: an instance of
                :class:`henson.base.Application` and the incoming
                message. It will be called for each incoming message
                with its result being passed to ``callback``.

        Returns:
            asyncio.coroutine: The callback.

        Raises:
            TypeError: If the callback isn't a coroutine.
        """
        self._register_callback(callback, 'message_preprocessor')
        return callback

    def result_postprocessor(self, callback):
        """Register a result postprocessing callback.

        Args:
            callback (asyncio.coroutine): A callable object that takes
                two arguments: an instance of
                :class:`henson.base.Application` and a result of
                processing the incoming message. It will be called for
                each result returned from ``callback``.

        Returns:
            asyncio.coroutine: The callback.

        Raises:
            TypeError: If the callback isn't a coroutine.
        """
        self._register_callback(callback, 'result_postprocessor')
        return callback

    def run_forever(self, num_workers=1, loop=None, debug=False):
        """Consume from the consumer until interrupted.

        Args:
            num_workers (Optional[int]): The number of asynchronous
                tasks to use to process messages received through the
                consumer.  Defaults to 1.
            loop (Optional[asyncio.asyncio.BaseEventLoop]): An event
                loop that, if provided, will be used for running the
                application. If none is provided, the default event loop
                will be used.
            debug (Optional[bool]): Whether or not to run with debug
                mode enabled. Defaults to True.

        Raises:
            TypeError: If the consumer is None or the callback isn't a
                coroutine.
        """
        if self.consumer is None:
            raise TypeError("The Application's consumer cannot be None.")

        if not asyncio.iscoroutinefunction(self.callback):
            raise TypeError("The Application's callback must be a coroutine.")

        # Use the specified event loop, otherwise use the default one.
        loop = loop or asyncio.get_event_loop()
        asyncio.set_event_loop(loop)

        # Start the application.
        tasks = [
            asyncio.async(callback(self), loop=loop) for callback in
            self._callbacks['startup']
        ]
        future = asyncio.gather(*tasks, loop=loop)
        loop.run_until_complete(future)

        # The following debug mode checks are intentionally separate.
        # Using a check of `if debug or self.settings['DEBUG']` would
        # accomplish the same thing but wouldn't respect the
        # PYTHONASYNCIODEBUG environment variable.
        if debug:
            # Set the application's debug mode to true if run_forever
            # was called with debug enabled.
            self.settings['DEBUG'] = True
        if self.settings['DEBUG']:
            # If the application is running in debug mode, enable it for
            # the loop and set the logger to DEBUG.
            loop.set_debug(True)
            self.logger.setLevel(logging.DEBUG)

        self.logger.info('application.started')

        # Create an asynchronous queue to pass the messages from the
        # consumer to the processor. The queue should hold one message
        # for each processing task.
        queue = asyncio.Queue(maxsize=num_workers, loop=loop)

        # Create a future to control consumption and create a task for
        # the consumer to run in.
        consumer = asyncio.Future(loop=loop)
        loop.create_task(self._consume(queue, consumer))

        # Create tasks to process each message received by the
        # consumer and wrap them inside a future. When the loop stops
        # running it should be restarted and wait until the future is
        # done.
        tasks = [
            asyncio.async(self._process(consumer, queue, loop), loop=loop)
            for _ in range(num_workers)
        ]
        future = asyncio.gather(*tasks, loop=loop)

        try:
            # Run the loop until the consumer says to stop.
            loop.run_until_complete(consumer)
        except BaseException as e:
            self.logger.error(e)

            # If something went wrong, cancel the consumer. This will
            # alert the processors to stop once the queue is empty.
            consumer.cancel()
        finally:
            # Run the loop until the future completes. This will allow
            # the tasks to finish processing all of the messages in the
            # queue and then exit cleanly.
            loop.run_until_complete(future)

            # Check for any exceptions that may have been raised by the
            # tasks inside the future.
            exc = future.exception()
            if exc:
                self.logger.error(exc)

            # Teardown
            tasks = [
                asyncio.async(callback(self), loop=loop) for callback in
                self._callbacks['teardown']
            ]
            future = asyncio.gather(*tasks, loop=loop)
            loop.run_until_complete(future)

            # Clean up after ourselves.
            loop.close()

        self.logger.info('application.stopped')

    def startup(self, callback):
        """Register a startup callback.

        Args:
            callback (asyncio.coroutine): A callable object that takes
                an instance of :class:`~henson.base.Application` as its
                only argument. It will be called once when the
                application first starts up.

        Returns:
            asyncio.coroutine: The callback.

        Raises:
            TypeError: If the callback isn't a coroutine.
        """
        self._register_callback(callback, 'startup')
        return callback

    def teardown(self, callback):
        """Register a teardown callback.

        Args:
            callback (asyncio.coroutine): A callable object that takes
                an instance of :class:`~henson.base.Application` as its
                only argument. It will be called once when the
                application is shutting down.

        Returns:
            asyncio.coroutine: The callback.

        Raises:
            TypeError: If the callback isn't a coroutine.
        """
        self._register_callback(callback, 'teardown')
        return callback

    @asyncio.coroutine
    def _abort(self, exc):
        """Log the aborted message.

        Args:
            exc (henson.exceptions.Abort): The exception to be logged.
        """
        tb = sys.exc_info()[-1]
        stack = traceback.extract_tb(tb, 1)[-1]
        self.logger.info('callback.aborted', extra={
            'exception': exc,
            'exception_message': exc.message,
            'aborted_by': stack,
        })

    @asyncio.coroutine
    def _apply_callbacks(self, callbacks, value):
        """Apply callbacks to a set of arguments.

        The callbacks will be called in the order in which they are
        specified, with the return value of each being passed to the
        next callback.

        Args:
            callbacks (List[callable]): The callbacks to apply to the
                provided arguments.
            value: The value to pass to the first callback.

        Returns:
            The return value of the final callback.
        """
        for callback in callbacks:
            value = yield from callback(self, value)
        return value

    @asyncio.coroutine
    def _consume(self, queue, future):
        """Read in incoming messages.

        Messages will be read from the consumer until it raises an
        :class:`~henson.exceptions.Abort` exception.

        Args:
            queue (asyncio.Queue): Any messages read in by the consumer
                will be added to the queue to share them with any future
                processing the messages.
            future (asyncio.Future): When the consumer tells the
                application to stop, this future will be cancelled.
        """
        while True:
            # Read messages and add them to the queue.
            try:
                value = yield from self.consumer.read()
            except Abort:
                self.logger.info('consumer.aborted')
                future.cancel()
                return
            except Exception as e:
                # If the consumer fails, set the exception on the future
                # so that the loop will stop running and the application
                # will shut down.
                future.set_exception(e)
            else:
                yield from queue.put(value)

    @asyncio.coroutine
    def _process(self, task, queue, loop):
        """Process incoming messages.

        Args:
            task (asyncio.tasks.Task): The task populating ``queue``.
              The function will exit when it's been cancelled.
            queue (asyncio.Queue): A queue containing incoming messages
              to be processed.
            loop (asyncio.asyncio.BaseEventLoop): The event loop used by
                the application.
        """
        while True:
            if queue.empty():
                # If there aren't any messages in the queue, check to
                # see if the consumer is done. If it is, exit.
                # Otherwise yield control back to the event loop and
                # then try again.
                if task.done():
                    break

                yield from asyncio.sleep(
                    self.settings['SLEEP_TIME'], loop=loop)
                continue

            message = yield from queue.get()
            # Save a copy of the original message in case its needed
            # later.
            original_message = deepcopy(message)

            try:
                message = yield from self._apply_callbacks(
                    self._callbacks['message_preprocessor'], message)
                self.logger.info('message.preprocessed')

                results = yield from self.callback(self, message)
            except Abort as e:
                yield from self._abort(e)
            except Exception as e:
                self.logger.error('message.failed', exc_info=sys.exc_info())

                for callback in self._callbacks['error']:
                    # Any callback can prevent execution of further
                    # callbacks by raising Abort.
                    try:
                        yield from callback(self, message, e)
                    except Abort:
                        break
            else:
                yield from self._postprocess_results(results)
            finally:
                # Don't use _apply_callbacks here since we want to pass
                # the original message into each callback.
                for callback in self._callbacks['message_acknowledgement']:
                    yield from callback(self, original_message)
                self.logger.info('message.acknowledged')

    @asyncio.coroutine
    def _postprocess_results(self, results):
        """Postprocess the results.

        Args:
            results (iterable): The results returned by processing the
                message.
        """
        if results is None:
            return

        for result in results:
            try:
                yield from self._apply_callbacks(
                    self._callbacks['result_postprocessor'], result)
                self.logger.info('result.postprocessed')
            except Abort as e:
                yield from self._abort(e)

    def _register_callback(self, callback, callback_container):
        """Register a callback.

        Args:
            callback (asyncio.coroutine): The callback to register.
            callback_container (str): The name of the container onto
                which to append the callback.

        Raises:
            TypeError: If the callback isn't a coroutine.
        """
        if not asyncio.iscoroutinefunction(callback):
            raise TypeError('The callback must be a coroutine.')

        self._callbacks[callback_container].append(callback)

        self.logger.info('callback.registered', extra={
            'type': callback_container,
            'callback': callback.__qualname__,
        })

    def _teardown(self, future, loop):
        """Tear down the application."""
        tasks = [
            asyncio.async(callback(self), loop=loop) for callback in
            self._callbacks['teardown']]
        future = asyncio.gather(*tasks, loop=loop)
        loop.run_until_complete(future)
